# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

# Import common lower level function for backwards compatibility
from ..utils.unitcell import molar_volume_from_unit_cell_volume
